/* mbRFslv3.c
   
   UART to PC via RF - Sends unsolicited data packets to RF master.
   
   One way comm only - does not receive data or requests from master.

Rev History ***************************************   
   
   Rev. 0  Built from mbRFslv.c modules
           Untested
   
   
   by Don Carveth 5/2002 - Program may be freely used and distributed
           for noncommercial use.  Don Carveth, or Botgoodies, make no
           warranties or guarantees regarding the ability of this program
           to perform its stated function.  I would not use this software
           to control your grandmothers heart machine. 
Acronyms *******************************************

	PC - PC Computer or microcontroller - device being communicated with
          via UART0
	5050 Character - A subset of the ASCII character set having about 50% 1s and 0s
             distributed more or less evenly

Program Operation **********************************

Routines used to send a data return packet using Mailbox protocol.  The
receiving master must be expecting a packet of the same fixed length.

The packet format is:           ::AaDdDdDd....Xx

      where
          :  is the start character - signifies start of a valid transmission
          Aa are the MS nibble and LS nibble of the Mailbox Address byte
          Dd are the MS nibble and LS nibble of the data bytes
          Xx are the MS nibble and LS nibble of the checksum byte 

Transmitter radio power must be on while transmitting in this mode. 
The Message in Progress LED flashes during message sends.   
The LED may be deleted or assigned to a different I/O location.

A timeout is included which resets all of the pointers if no activity has occured
after a preset timeout time.  The timer is reset each send / receive and is monitored
in the main program loop or checked by a timer ISR.  The ideal timeout value should
be about 4 times the duaration of a byte which varies with baudrate.  A single
character at 9600 baud has a duration of about 1 ms.  At 9600 baud then
the timeout would be set at about 4 ms, at 2400 baud 16 ms. Actually any timeout
amount up to even seconds is OK if you can afford to have the routines 
non-functional for that long.  Setup the RESET_TIMER parameters in mbRFslv3.h
based on the reset method used and call CheckRFTimeout() from the main program
periodically.



*/
#include <stdlib.h>
#include <io.h>
#include <sig-avr.h>
#include <interrupt.h>
#include <string-avr.h>

#include "StdDefs.h"
#include "mbRFslv.h"

#define RCV_ADDR    RFRcvBuff[2]
#define RCV_LEN    (RFRcvBuff[3] >> 3)
#define RCV_RORW   (RFRcvBuff[3] & 0x01)

const CHARU F50Table[23] = {0x3C,0x47,0x4B,0x33,0x4E,0x35,0x36,0x59,0x5A,0x39,
                    0x63,0x65,0x66,0x69,0x6A,0x6C,0x71,0x72,0x74,0x78,
                    0x4D,0x53,0x56}; 

//UART Xmt Buffer holds byte data before 5050 /AHex conversion
volatile CHARU RFXmtBuff[MAX_BYTES];

CHARU hex_to_char(CHARU MSB,  CHARU LSB);
CHARU char_to_hex(CHARU c);
void HA_To5050(CHARU *pack, INT16S LengthLess1);
CHARU HAC_To5050(CHARU C);
CHARU Fifty50ToHA(CHARU C);

volatile CHARU MP;          // UART xmt data pointer
volatile CHARU Checksum;		//Used to hold calculated checksum value
volatile CHARU CS;          // Checksum as calculated
volatile CHARU XmtNumDataBytes;  // Number of data bytes to xmt from UART
volatile CHARU MBXmtFlag;   // UART to transmit flag
INT16U OneByteDelayTime;   // Used in MSleep as 1/10s of a second

CHARU *BufPtr;                   

/******************** USER SETUP *******************************/
void InitializeRF(void)
   {
   // Set up UART           
   outp(BAUDRATE, UART_BAUD_REG);
   outp(0x18, UART_CONTROL_REG); // B8 - Tx/Rx enable, Interrupt: Rx/DRE Interrupt enabled
		  					 // 18 - Tx/Rx enable, Interrupt: Rx/DRE Interrupt disabled
							 // 98 - Tx/Rx enable, Interrupt: Rx enabled, DRE disabled 
   OneByteDelayTime = 960/(BAUD/100);
   }

    
// Interrupt Driven UART Transmit Routine
SIGNAL(SIG_UART_DATA)
//SIGNAL(UART_REG_EMPTY_INT_VECTOR)
    {
    RF_LED_ON();
    RESET_TIMER();
    
    outp(RFXmtBuff[MP], UART_DATA_REG);
    if (MP < XmtNumDataBytes)
       {
      	MP++;
       }
  	 else             // Last byte
       {
      	MP = 0;
      	MBXmtFlag = 0;
        outp(0x18, UART_CONTROL_REG);  // Disable UART Transmit ISR
        msleep(OneByteDelayTime);
        msleep(OneByteDelayTime);
        msleep(OneByteDelayTime/2);
        RF_XMT_POWER_OFF();
        //RF_RCV_INTERRUPT_ENABLED();
      	}
    }    

// Send data to an MB master.  Can be called directly - unsolicited data, or
// is called by ReadMB as the reply to a master read request.
void SendToMaster(CHARU addr, CHARU len, CHARU *Data)
   {   
   CHARU j,k;
   CHARU CS_calc;

   // Turn on RF xmt power
   RF_XMT_POWER_ON();
   RF_RCV_INTERRUPT_DISABLED();
   msleep(20);
   
   // Preamble
   RFXmtBuff[0] = 0x55;   // Sync characters
   RFXmtBuff[1] = 0x55;
   RFXmtBuff[2] = 0x55;
   RFXmtBuff[3] = 0x55;
   RFXmtBuff[4] = 0xFF;
   RFXmtBuff[5] = 0x63;   // ":"  63
   RFXmtBuff[6] = 0x63;   // ":"
   // Address
   RFXmtBuff[7] = HAC_To5050(char_to_hex(addr>>4));
   RFXmtBuff[8] = HAC_To5050(char_to_hex(addr & 0x0F));
   CS_calc = addr;
   // Data
   k = 9;
   for (j = 0; j < (len); j++)
   {
   RFXmtBuff[k] = HAC_To5050(char_to_hex((Data[j])>>4));
   RFXmtBuff[k+1] = HAC_To5050(char_to_hex((Data[j]) & 0x0F));
   k+=2;
   CS_calc += Data[j];
   }
   // Add checksum
   RFXmtBuff[k] = HAC_To5050(char_to_hex(CS_calc>>4));
   RFXmtBuff[k + 1] = HAC_To5050(char_to_hex(CS_calc & 0x0F));

   // Control now transfers to the Xmt ISR
   outp(RFXmtBuff[0], UART_DATA_REG);   
	outp(0x38, UART_CONTROL_REG);	// Enable UART Data Register Empty ISR
   MBXmtFlag = 1;
   XmtNumDataBytes = (2*len) + 10;
   MP = 1;
   }
   
// Resets the UART ISR pointers if a message is hung up
void CheckRFTimeout(void)
     {
     if ((RESET_TIMER_REG - RFTimeout) > TIMEOUT_PERIOD)
       {
       outp(0x98, UART_CONTROL_REG);	 // Disable UART Transmit ISR
       MBXmtFlag = 0;
	    // PORTC &= ~0x02;
       RESET_TIMER();
       RF_XMT_POWER_OFF();
       }
     }
     
CHARU hex_to_char(CHARU MSB,  CHARU LSB)
	{
	INT16S intval;

	if (LSB > 0x2Fu)
		LSB = LSB - 0x30u;
	else
		LSB = 0u;
	if (LSB > 0x9u)
		LSB = LSB - 7;

	if (MSB > 0x2Fu)
		MSB = MSB - 0x30u;
	else
		MSB = 0u;
	if (MSB > 0x9u)
		MSB = MSB - 7;
	MSB = MSB * 16;

	intval = LSB + MSB;
	return intval;
	}

CHARU char_to_hex(CHARU c)	/* Returns ASCII value for 0-F */
	{
	if (c < 10) c = c + 0x30;
		else c = c + 0x37;
	return c;
	}

void HA_To5050(CHARU *pack, INT16S LengthLess1)
    {
    INT16S j;
    for (j = 0; j <= LengthLess1; j++)
           *(pack + j) = F50Table[*(pack + j) - 0x30];
    }
    
CHARU HAC_To5050(CHARU C)
    {
    return F50Table[C - 0x30];
    }    

CHARU Fifty50ToHA(CHARU C)
   {
   int k;
   CHARU C1 = '@';
   for (k = 0; k < 23; k++)         /* Locate char in 5050 table */
     {
     if (C == F50Table[k])
        {
        C1 = k + 0x30;
        break;
        }
     }
   return C1;
   } 
   